import java.util.*;
import java.io.*;

public class DFA {
    static int Nmax=2285760;  //an upper bound for the nb of states
    Map<Integer, Integer> next[];          // the nextstate function
    Integer initial;           // the initial state
    Set<Integer> terminal;    // the partition of terminal states
    int nbStates;

    /**
       creates a DFA with n states and alphabet a.
    */
    public   DFA(int n) {
	nbStates = 0; 
	next = new TreeMap[n];
	terminal = new TreeSet<Integer>();
    }
    
     public void SetTerminal(int i) {
	terminal.add(i);
    }
    int next(int p, short[] w){
	//transition by a word   w in a DFA
	Integer q = new Integer(p);
	for(int i = 0; i < w.length; i++){
	    q = next[q].get(w[i]);
	    if (q == null) break;
	}
	return q;
    }
    public void addEdge(int p, int c, int q) {
	//	if (next[p].containsKey(c)) {
	//	    System.out.println("erreur : arte dj prsente");
	//	}
	//	else {
	    next[p].put(c, q);
	    //	}
    }
    public int newState() {
	int n = nbStates;
	next[n] = new TreeMap();
	nbStates++;
	return n;
    }
    public NFA toNFA() {
       NFA a = new NFA(nbStates);
	
	for (int i = 0; i < nbStates; i++) {
	    a.newState();
	    for (Integer c : next[i].keySet()) {
		a.addEdge(i, c, next[i].get(c));
	    }
	}
	a.terminal = new TreeSet<Integer>(terminal);
	a.initial = new TreeSet<Integer>();
	a.initial.add(initial);
	return a;
    }
    public NFA reverse() {
	NFA a = toNFA();
	return a.reverse();
    }

    void addWord(String w) {
	if (initial == null) {
	    int n = newState();
	    initial = n;
	}
	
	int i = 0;
	Integer p = i;
	while (i < w.length()) {
	    int c = (int) w.charAt(i);
	    if (!next[p].containsKey(c)) {
		int n = newState();
		addEdge(p, c, n);
		p = n;
	    }
	    else {
		p = next[p].get(c);
	    }
	    i++;
	}
	SetTerminal(p);
    }
	    
    public String toString(){
	String s = "initial = " + initial+"\n" + "terminals = " + terminal;
	s+="nbStates = "+nbStates+"\n";
	s+="  ";
	s+="\n";
	StringBuffer u = new StringBuffer(s);
	for(int i=0;i<nbStates;i++){
	    u.append("** State " + i+"\n");
	    for(Integer c : next[i].keySet()) {
		u.append("<"+c+"> : " + next[i].get(c)+"\n");
	    }
	}
	s = u.toString();
	return s;
    }
    public static DFA fromFile(String name) throws IOException {
	FileReader fileIn = new FileReader(name);
	BufferedReader r = new BufferedReader(fileIn);
	String line;	
	int count=0;
	DFA a = new DFA(Nmax);
	while ((line = r.readLine()) != null) {
  	    count++;
  	    if(count % 10000 == 0){		
		System.out.println("nbWords = " + count);
	    }
	    a.addWord(line);
	}
	fileIn.close();
	System.out.println("Total  = " + count);
	System.out.println("Number of states  = " + a.nbStates);
	return a;
    }

}
